"""
The actual server
"""
import logging
import socket
import os
import sys
import select
import WebFrame as WF

class __LRU:
    class Node:
        def __init__(self, data):
            self.next = None
            self.prev = None
            self.data = data

        def __str__(self):
            return "Node: "+str(self.data)+"\n"+str(self.next)
            
            
    def __init__(self):
        self.__lru_top = None
        self.__lru_bottom = None
        self.__keys = {}

    def __str__(self):
        return str(self.__lru_bottom)

    def access(self, key):
        logging.debug("LRU access [%s]", key)
        logging.debug(str(self))

        if key in self.__keys:
            node = self.__keys[key]
        else:
            logging.debug("LRU new [%s]", key)
            node = self.Node(key)
            self.__keys[key] = node

        # Remove from list
        if node.next: node.next.prev = node.prev
        if node.prev: node.prev.next = node.next
        if self.__lru_top == node: self.__lru_top = node.prev
        if self.__lru_bottom == node: self.__lru_bottom = node.next
        node.next = None
        node.prev = None

        # Insert at top
        node.prev = self.__lru_top
        if self.__lru_top: self.__lru_top.next = node
        self.__lru_top = node

        # Check bottom
        if self.__lru_bottom == None:
            self.__lru_bottom = node

    def pop(self):
        if self.__lru_bottom == None:
            return None

        node = self.__lru_bottom
        logging.debug("LRU removed [%s]", node.data)
        logging.debug(str(self))

        # Remove from list
        self.__lru_bottom = node.next
        if node.next: node.next.prev = None

        del self.__keys[node.data]

        return node.data

class __Req:
    def __init__(self, sock, what):
        logging.debug("New request %s", what)
        self.state = "keylen"
        self.sock = sock
        self.what = what
        self.keylen = None
        self.vallen = None
        self.key = None
        self.val = None
        
    def read(self):
        if self.state == "keylen":
            self.keylen = int(self.sock.recv(4))
            logging.debug("Req(%s) keylen=%d", self.what, self.keylen)
            if self.what == "S": self.state = "vallen"
            else: self.state = "getkey"
            return False

        elif self.state == "vallen":
            self.vallen = int(self.sock.recv(4))
            logging.debug("Req(%s) vallen=%d", self.what, self.vallen)
            self.state = "getkey"
            return False

        elif self.state == "getkey":
            self.key = self.sock.recv(self.keylen)
            if self.what == "S": self.state = "getval"
            else:
                self.get()
                return True
            
        elif self.state == "getval":
            self.val = self.sock.recv(self.vallen)
            self.set()
            return True

        else:
            logging.error("Unknown state %s", self.state)
            return True

    def get(self):
        global _cache, _lru
        logging.debug("Req(%s): Getting [%s]", self.what, self.key)
        
        if self.key in _cache:
            _lru.access(self.key)
            val = _cache[self.key]
            self.sock.send("%04d%s" % (len(val), val))
        else:
            self.sock.send("0000")

    def set(self):
        global _cache, _lru
        logging.debug("Req(%s): Setting [%s] -> [%s]", self.what, self.key,
                      self.val)
        _lru.access(self.key)
        _cache[self.key] = self.val

        if len(_cache) > WF.get_opt("cache.entries"):
            key = _lru.pop()
            del _cache[key]





# Global storage
_cache = {}
_lru = __LRU()


def __server():
    reqs = {}
    clients = []
    
    # Setup our socket
    sname = WF.get_opt("cache.sockname") % WF.get_opt("app.name")
    try:
        os.unlink(sname)
    except OSError:
        pass
    s = socket.socket(socket.AF_UNIX)
    s.bind(sname)
    s.listen(15)
    
    # The main server loop
    while True:
        slist = [s] + clients
        res = select.select(slist, [], [], 300)
        for sock in res[0]:
            if sock == s:
                con = s.accept()
                logging.debug("New connection")
                clients.append(con[0])
            else:
                if sock in reqs:
                    try:
                        if reqs[sock].read():
                            del reqs[sock]
                    except socket.error:
                        logging.debug("1-Connection broken with %s", sock)
                        clients.remove(sock)
                else:
                    try:
                        what = sock.recv(1)
                        if len(what) == 0:
                            logging.debug("3-Connection broken with %s", sock)
                            clients.remove(sock)                            
                        elif what == "G" or what == "S":
                            reqs[sock] = __Req(sock, what)
                    except socket.error:
                        logging.debug("2-Connection broken with %s", sock)
                        clients.remove(sock)


def __start():
    # Fork off 
    if os.fork():
        return

    logging.debug("Cacheserver first fork")
    logging.shutdown()
    
    # Close all fds
    for i in range(0, 100):
        try:
            os.close(i)
        except OSError:
            pass

    # Second fork, to totally remove us from parent.
    if os.fork():
        sys.exit()
    try:
        os.setsid()
    except OSError:
        pass
    try:
        os.setpgrp()
    except OSError:
        pass

    # Override logging
    reload(logging)
    lname = WF.get_opt("cache.logfile") % WF.get_opt("app.name")
    level = WF.get_opt("logging.level")
    if level == None:
        if WF.get_opt("debug"): level = logging.DEBUG
        else: level = logging.INFO
    logging.basicConfig(filename = lname,
                        level = level,
                        format = WF.get_opt("logging.format"))
    logging.debug("Cache server started with pid %d", os.getpid())

    # Run the server
    try:
        __server()
    except Exception, e:
        logging.error(str(e))
        raise

