"""
User authentification and permissions module.

This module contains a User class and also functions to perform user
authentification and permission verification.

The module contains helpers for authentification and user retrieval based
on a simple user object.

The module also implements a access policy checker to determine if the user
may access a resource based on the policies attached to that resource.
Each resource has two policies attached: A read policy and a write policy.
These policies determine how the resource may be accessed.
The syntax of the policies is simple, they are a string containing a
comma(,)-separated list of tags identifying rights. The tags are
tested in order of appearance, from left to right. 

Some special tags are recognized:
    @deny       - Deny to all
    @tls        - Require that the connection is TLS/SSL encrypted
    @loginpage  - Send a login page if user not logged in
    @loginredir - Redirect to other URL if not logged in
    @login401   - Send a HTTP 401 response if user not logged in
    @public     - Access is granted to anyone
    @loggedin   - Access is granted to logged in users.
    @superuser  - Access is granted to superusers.

Some example policies:
    policy = "@public"
        Access is allowed to anyone requesting this object
    policy = "accounting,boss,janitor"
        Access is granted to any user that has one or more of the tags
        "accounting", "boss" or "janitor" associated with them.
    
"""
import logging
import WebFrame as WF

readonly_methods = ["GET", "HEAD"]

class User(object):
    """The user. Keeps track of username and superuser status."""

    def __init__(self, user, superuser=False):
        self.user = user
        self.superuser = superuser
        self.params = {}
        self.__permissions = {}

    def set(self, param, value):
        """Set a parameter related to a user.
        These are usually set from the database when the user is read.
        Setting them here won't change the database!"""
        self.params[param] = value

    def get(self, param):
        """Get a parameter related to a user.
        These are usually set from the database when the user is read."""
        return self.params[value]

    def add_permission(self, tag):
        """Add a permission tag to the user object.
        These are matched literally and case-sensitively against the
        auth policy when user acces is checked"""
        self.__permissions[tag]

    def check_permission(self, tag):
        """Check if this user has been granted a particular permission.
        Superusers are automatically granted all permissions."""
        if self.superuser:
            return True
        else:
            return tag in self.__permissions


def get_user(user, passwd):
    """This function tries to authenticate the user and if it succeeds
    it returns a User object with the user data.
    The parameters in the object are taken from the sql query provided in
    'auth.get_auth_user_sql'.
    The column names provided in 'auth.remove_user_colums' are removed.
    
    """
    
    dbh = WF.open_db()
    cr = dbh.simple(WF.get_opt("auth.get_auth_user_sql"), { "user" : user,
                                                            "passwd" : passwd})
    if cr.rowcount != 1:
        logging.info("Auth FAILED for user '%s'", user)
        return None
    else:
        logging.info("Auth OK for user %s", user)
        ud = cr.fetchone()
        u = User(user)
        if "superuser" in ud and ud["superuser"]:
            u.superuser = True

        # Remove password, try different names
        for i in WF.get_opt("auth.remove_user_colums"):
            if i in ud:
                del ud[i]
                if WF.get_opt("debug"):
                    logging.debug("Removed '%s' from user object", i)
        
        u.params = ud

        return u

def _ask_for_login(ctx):
    logging.debug("Ask-pw method is '%s'", ctx._login_meth)
    if ctx._login_meth == "login401":
        return ctx.resp.login_required()
    elif ctx._login_meth == "loginpage":
        from WebFrame import template
        if WF.get_opt("debug"):
            logging.debug("Sending login page '%s'",
                          WF.get_opt("auth.loginpage_path"))
            ctx.resp.write(
                template.run_template(ctx, WF.get_opt("auth.loginpage_path")))
        return ctx.resp.stop_processing()
    elif ctx._login_meth == "loginredir":
        if WF.get_opt("debug"):
            logging.debug("Redirecting to login page at '%s'",
                          WF.get_opt("auth.loginredir_uri"))
        ctx.resp.redir(WF.get_opt("auth.loginredir_uri"))
        return ctx.stop_processing()
    else:
        raise ValueError("Bad login request specification")

def _parse_policy(policy):
    policy = policy.split(",")
    for i in policy:
        i.strip()
    return policy

def check_policy(ctx, policy):
    """Check if a policy specification allows access to this resource.
    If you use this function please take care to test both for read and
    write access as appropriate for your function.
    Returns:
        True  if policy allows access,
        False if it doesn't (because user not logged in or no rights)
        None  if the policy cannot allow access at all for this connection.
    """
    if policy == None:
        return False
    policy = _parse_policy(policy)
    ctx._login_meth = WF.get_opt("auth.default_login_method")
    
    # Loop through specifiers in order
    for i in policy:
        if i == "@deny":
            return None
        elif i == "@tls":
            if not ctx.req.tls:
                return None
        elif i == "@login401":
            ctx._login_meth = "login401"
        elif i == "@loginpage":
            ctx._login_meth = "loginpage"
        elif i == "@loginredir":
            ctx._login_meth = "loginredir"
        elif i == "@public":
            # For public resources
            return True
        elif i == "@loggedin":
            if ctx.user:
                # Logged in users allowed, and logged in user found.
                return True
        elif i == "@superuser":
            if ctx.user and ctx.user.superuser:
                # Superusers allowed, and superuser is logged in
                return True
        elif ctx.user and ctx.user.check_permission(i):
            # Check if user has specific permission tag
            return True
        else:
            return False

    # Nothing matched, so false it is.
    return False

def check_access(ctx, read, write):
    """Check if the policies allow access for this object.
    This function considers the request method and two policy
    specifications when making its decision."""
    global readonly_methods
    if ctx.req.method in readonly_methods:
        if WF.get_opt("debug"):
            logging.debug("Checking read policy '%s' for method %s",
                          read, ctx.req.method)
        policy = read
    else:
        if WF.get_opt("debug"):
            logging.debug("Checking write policy '%s' for method %s",
                          write, ctx.req.method)
        policy = write

    cp = check_policy(ctx, policy)
    if cp == None:
        # No way, never, not a chance
        if WF.get_opt("debug"):
            logging.debug("Policy doesn't allow any access at all!")
        ctx.resp.access_denied()
        return False
    elif cp == False:
        if WF.get_opt("debug"):
            logging.debug("User unauthorized / not logged in")
        if ctx.user:
            ctx.resp.access_denied()
        else:
            _ask_for_login(ctx)
        return False
    elif cp == True:
        if WF.get_opt("debug"):
            logging.debug("Access granted")
        return True
    else:
        # Should never happen
        raise ValueError
    

def _login_helper(ctx, action):
    login = None
    pwd = None
    if "login" in ctx.req.params: login = ctx.req.params["login"][0]
    if "password" in ctx.req.params: pwd = ctx.req.params["password"][0]
    if login != None and pwd != None:
        u = get_user(login, pwd)
        ctx._set_user(u)
    return ctx.resp.cont()

def _logout_helper(ctx, action):
    if WF.get_opt("debug"):
        if ctx.user != None:
            logging.debug("Logout for '%s'", ctx.user.user)
    ctx._clear_user()

def _init():
    from WebFrame import handler
    if WF.get_opt("auth.form_login_helpers"):
        handler.register_query("LOGIN", _login_helper, "@public", "@public")
        handler.register_query("LOGOUT", _logout_helper, "@public", "@public")
        
