"""
WebFrame data cache system.

Provides a simple cache for data that is to be shared between
requests. No synchronization is provided so it must be used in such a
manner that races aren't a problem.

The cache runs in its own process and provides a local unix socket so
that WebFrame may query it.
If the cache server is not running then it will get started
automatically by WebFrame on the first request. The daemon runs with
the same user/group and permissions as WebFrame.
So you do not need to run any daemons yourself.

WebFrame currently uses the cache to store language translation data
and session data

The cache is organized into "sections". The sections are nothing more
than arbitrary strings, and you may use them as you wish.
Just don't use any sections starting witn "__WF_" since those are used
internally by WebFrame.
"""
import logging
import socket
import os
import sys
import cPickle
import WebFrame as WF
import _cacheserver

__local_cache = {}
__socket = None
__starting_cache = False

# Temporary storage from has, so we don't do two requests
__has_val = None
__has_key = None

def __connect(startserver=False):
    global __socket, __starting_cache
    
    # Check if we are connected and if cache should be used
    if __socket != None or not WF.get_opt('cache.active'):
        return

    # Connect to cache
    sname = WF.get_opt("cache.sockname") % WF.get_opt("app.name")
    s = socket.socket(socket.AF_UNIX)
    try:
        s.connect(sname)
    except socket.error:
        # Cache not active, try to start it
        if startserver and not __starting_cache:
            logging.debug("Starting cache server")
            __starting_cache = True
            _cacheserver.__start()
        return

    # Connected, set socket
    __socket = s
    
    return

def __query_cache(key):
    global __local_cache, __socket

    # Check in local cache
    if key in __local_cache:
        logging.debug("Local cache hit")
        return __local_cache[key]

    # Query the server
    if __socket:
        try:
            __socket.send("G%04d" % len(key))
            __socket.send(key)
        except socket.error:
            logging.debug("Connection to cache broken")
            __socket = None
            raise KeyError
        
        l = int(__socket.recv(4))
        if l == 0:
            raise KeyError
        else:
            val = __socket.recv(l)
            val = cPickle.loads(val)
        return val
    else:
        raise KeyError


def has(section, key):
    """Check if the cache contains a specified key.
    """
    global __has_key, __has_value
    __connect()
    key = unicode(section)+":"+unicode(key)

    # First check local cache for immutables
    try:
        val = __query_cache(key)
    except KeyError:
        return False
    else:
        __has_key = key
        __has_value = val
        return True


def get(section, key):
    """Request a key from the cache.
    """
    global __has_key, __has_value
    __connect()
    key = unicode(section)+":"+unicode(key)

    # Check the temporary has-space if the user just did a has() 
    if key == __has_key:
        __has_key = None
        return __has_value

    # Clear out has
    __has_key = None
    __has_value = None

    return __query_cache(key)

def set(section, key, value, local=False):
    """local - True if this value may be cached locally. Only set it
    for things that are not changed between requests
    """
    global __has_key, __has_value, __local_cache, __socket, __file
    __connect()
    key = unicode(section)+":"+unicode(key)
    
    # Clear out has
    __has_key = None
    __has_value = None

    # Set in local cache
    if local:
        __local_cache[key] = value
    
    # Send to cache server
    if __socket != None:
        pic = cPickle.dumps(value, -1)
        try:
            __socket.send("S%04d%04d" % (len(key), len(pic)))
            __socket.send(key)
            __socket.send(pic)
        except socket.error:
            logging.debug("Connection to cache broken on set")
            __socket = None

